"""

Projection Information Source: https://developers.arcgis.com/rest/services-reference/projected-coordinate-systems.htm
Geographic Projection Source:
"""
import os
import json
import pandas as pd

__all__ = ['lookup_coordinate_system', 'info']
#----------------------------------------------------------------------
_lutbl = None
#----------------------------------------------------------------------
def _load_data():
    """loads the coordinate information into memory"""

    _fp = r"%s\prj.json" % os.path.dirname(__file__)
    global _lutbl
    if _lutbl is None:
        with open(_fp, 'r') as reader:
            _lutbl = pd.DataFrame(json.loads(reader.read()))
            _lutbl.columns = ['WKID', 'NAME', 'WKT']
            del reader
    return _lutbl
#----------------------------------------------------------------------
def lookup_coordinate_system(wkid):
    """
    Looks up the wkt and name of the wkid coordinate system

    ================     =========================================
    ** Arguement **      ** Description **
    ----------------     -----------------------------------------
    wkid                 Requred Integer/List. The SRS identifier.
    ================     =========================================

    :returns: list of dictionaries

    [
      {
         'WKID': 2000,
         'NAME': 'Anguilla_1957_British_West_Indies_Grid',
         'WKT': 'PROJCS["Anguilla_1957_British_West_Indies_Grid",GEOGCS["GCS_Anguilla_1957",DATUM["D_Anguilla_1957",SPHEROID["Clarke_1880_RGS",6378249.145,293.465]],PRIMEM["Greenwich",0.0],UNIT["Degree",0.0174532925199433]],PROJECTION["Transverse_Mercator"],PARAMETER["False_Easting",400000.0],PARAMETER["False_Northing",0.0],PARAMETER["Central_Meridian",-62.0],PARAMETER["Scale_Factor",0.9995],PARAMETER["Latitude_Of_Origin",0.0],UNIT["Meter",1.0]]'
      }
    ]


    """
    if wkid == 102100:
        wkid = 3857
    global _lutbl
    if _lutbl is None:
        _load_data()
    if isinstance(wkid, (int, float)):
        q = _lutbl['WKID'] == int(wkid)
    elif isinstance(wkid, (list, tuple)):
        q = _lutbl['WKID'].isin(wkid)
    else:
        raise ValueError("Invalid wkid. Must be int or list.")
    if len(_lutbl[q]) == 0:
        raise ValueError("Invalid WKID")
    return _lutbl[q].to_dict('records')
#----------------------------------------------------------------------
def info():
    """Returns all the support projections

    :returns: pd.DataFrame
    """
    global _lutbl
    if _lutbl is None:
        _lutbl = _load_data()
    return _lutbl



"""
This module contians operations to create, modify and update geopackages
"""
import os
import sys
import sqlite3
import tempfile
#--------------------------------------------------------------------------
_field_lookup = {
    "text" : ["TEXT", "check((typeof({field}) = 'text' or typeof({field}) = 'null') and not length({field}) > {l})"],
    "float" : ["DOUBLE", """check(typeof({field}) = 'real' or typeof({field}) = 'null')"""],
    "double" : ["DOUBLE", "check(typeof({field}) = 'real' or typeof({field}) = 'null')"],
    "short" : ["SMALLINT", "check((typeof({field}) = 'integer' or typeof({field}) = 'null') and {field} >= -32768 and {field} <= 32767)"],
    "long" : ["MEDIUMINT", "check((typeof({field}) = 'integer' or typeof({field}) = 'null') and {field} >= -2147483648 and {field} <= 2147483647)"],
    "integer" : ["MEDIUMINT", "check((typeof({field}) = 'integer' or typeof({field}) = 'null') and {field} >= -2147483648 and {field} <= 2147483647)"],
    "date" : ["DATETIME", "check((typeof({field}) = 'text' or typeof({field}) = 'null') and strftime('%Y-%m-%dT%H:%M:%fZ',{field}))"],
    "blob" : ["BLOB", "check(typeof({field}) = 'blob' or typeof({field}) = 'null')"],
    "guid" : ["TEXT(38)", "check((typeof({field}) = 'text' or typeof({field}) = 'null') and not length({field}) > 38)"]
}
#--------------------------------------------------------------------------
_create_tables_sql = [
    """CREATE TABLE IF NOT EXISTS gpkg_contents (table_name TEXT NOT NULL PRIMARY KEY,data_type TEXT NOT NULL,identifier TEXT UNIQUE,description TEXT DEFAULT '',last_change DATETIME NOT NULL DEFAULT (strftime('%Y-%m-%dT%H:%M:%fZ','now')),min_x DOUBLE,min_y DOUBLE,max_x DOUBLE,max_y DOUBLE,srs_id INTEGER,CONSTRAINT fk_gc_r_srs_id FOREIGN KEY (srs_id) REFERENCES gpkg_spatial_ref_sys (srs_id))""",
    """CREATE TABLE IF NOT EXISTS gpkg_data_column_constraints (constraint_name TEXT NOT NULL,constraint_type TEXT NOT NULL /* 'range' || 'enum' | 'glob' */,value TEXT,min NUMERIC,min_is_inclusive BOOLEAN,max NUMERIC,max_is_inclusive BOOLEAN,description TEXT,CONSTRAINT gdcc_ntv UNIQUE (constraint_name,constraint_type,value))""",
    """CREATE TABLE IF NOT EXISTS gpkg_data_columns (table_name TEXT NOT NULL,column_name TEXT NOT NULL,name TEXT,title TEXT,description TEXT,mime_type TEXT,constraint_name TEXT,CONSTRAINT pk_gdc PRIMARY KEY (table_name,column_name),CONSTRAINT fk_gdc_tn FOREIGN KEY (table_name) REFERENCES gpkg_contents (table_name))""",
    """CREATE TABLE IF NOT EXISTS gpkg_extensions (table_name TEXT,column_name TEXT,extension_name TEXT NOT NULL,definition TEXT NOT NULL,scope TEXT NOT NULL,CONSTRAINT ge_tce UNIQUE (table_name,column_name,extension_name))""",
    """CREATE TABLE IF NOT EXISTS gpkg_geometry_columns (table_name TEXT NOT NULL,column_name TEXT NOT NULL,geometry_type_name TEXT NOT NULL,srs_id INTEGER NOT NULL,z TINYINT NOT NULL,m TINYINT NOT NULL,CONSTRAINT pk_geom_cols PRIMARY KEY (table_name, column_name),CONSTRAINT uk_gc_table_name UNIQUE (table_name),CONSTRAINT fk_gc_tn FOREIGN KEY (table_name) REFERENCES gpkg_contents (table_name),CONSTRAINT fk_gc_srs FOREIGN KEY (srs_id) REFERENCES gpkg_spatial_ref_sys (srs_id))""",
    """CREATE TABLE IF NOT EXISTS gpkg_metadata (id INTEGER CONSTRAINT m_pk PRIMARY KEY ASC AUTOINCREMENT NOT NULL UNIQUE,md_scope TEXT NOT NULL DEFAULT 'dataset',md_standard_uri TEXT NOT NULL,mime_type TEXT NOT NULL DEFAULT 'text/xml',metadata TEXT NOT NULL)""",
    """CREATE TABLE IF NOT EXISTS gpkg_metadata_reference (reference_scope TEXT NOT NULL,table_name TEXT,column_name TEXT,row_id_value INTEGER,timestamp DATETIME NOT NULL DEFAULT (strftime('%Y-%m-%dT%H:%M:%fZ','now')),md_file_id INTEGER NOT NULL,md_parent_id INTEGER,CONSTRAINT crmr_mfi_fk FOREIGN KEY (md_file_id) REFERENCES gpkg_metadata(id),CONSTRAINT crmr_mpi_fk FOREIGN KEY (md_parent_id) REFERENCES gpkg_metadata(id))""",
    """CREATE TABLE IF NOT EXISTS gpkg_spatial_ref_sys (srs_name TEXT NOT NULL,srs_id INTEGER NOT NULL PRIMARY KEY,organization TEXT NOT NULL,organization_coordsys_id INTEGER NOT NULL,definition TEXT NOT NULL,description TEXT)""",
    """CREATE TABLE IF NOT EXISTS gpkg_tile_matrix (table_name TEXT NOT NULL,zoom_level INTEGER NOT NULL,matrix_width INTEGER NOT NULL,matrix_height INTEGER NOT NULL,tile_width INTEGER NOT NULL,tile_height INTEGER NOT NULL,pixel_x_size DOUBLE NOT NULL,pixel_y_size DOUBLE NOT NULL,CONSTRAINT pk_ttm PRIMARY KEY (table_name,zoom_level),CONSTRAINT fk_ttm_table_name FOREIGN KEY (table_name) REFERENCES gpkg_contents(table_name))""",
    """CREATE TABLE IF NOT EXISTS gpkg_tile_matrix_set (table_name TEXT NOT NULL PRIMARY KEY,srs_id INTEGER NOT NULL,min_x DOUBLE NOT NULL,min_y DOUBLE NOT NULL,max_x DOUBLE NOT NULL,max_y DOUBLE NOT NULL,CONSTRAINT fk_gtms_table_name FOREIGN KEY (table_name) REFERENCES gpkg_contents (table_name),CONSTRAINT fk_gtms_srs FOREIGN KEY (srs_id) REFERENCES gpkg_spatial_ref_sys (srs_id))"""
]
#--------------------------------------------------------------------------
_initial_triggers_sql = [
    """CREATE TRIGGER IF NOT EXISTS 'gpkg_metadata_md_scope_insert' BEFORE INSERT ON 'gpkg_metadata' FOR EACH ROW BEGIN SELECT RAISE(ABORT, 'insert on table gpkg_metadata violates constraint: md_scope must be one of undefined | fieldSession | collectionSession | series | dataset | featureType | feature | attributeType | attribute | tile | model | catalogue | schema | taxonomy | software | service | collectionHardware | nonGeographicDataset | dimensionGroup') WHERE NOT(NEW.md_scope IN ('undefined','fieldSession','collectionSession','series','dataset','featureType','feature','attributeType','attribute','tile','model','catalogue','schema','taxonomy','software','service','collectionHardware','nonGeographicDataset','dimensionGroup'));END""",
    """CREATE TRIGGER IF NOT EXISTS 'gpkg_metadata_md_scope_update' BEFORE UPDATE OF 'md_scope' ON 'gpkg_metadata' FOR EACH ROW BEGIN SELECT RAISE(ABORT, 'update on table gpkg_metadata violates constraint: md_scope must be one of undefined | fieldSession | collectionSession | series | dataset | featureType | feature | attributeType | attribute | tile | model | catalogue | schema | taxonomy | software | service | collectionHardware | nonGeographicDataset | dimensionGroup') WHERE NOT(NEW.md_scope IN ('undefined','fieldSession','collectionSession','series','dataset','featureType','feature','attributeType','attribute','tile','model','catalogue','schema','taxonomy','software','service','collectionHardware','nonGeographicDataset','dimensionGroup'));END""",
    """CREATE TRIGGER IF NOT EXISTS 'gpkg_metadata_reference_column_name_insert' BEFORE INSERT ON 'gpkg_metadata_reference' FOR EACH ROW BEGIN SELECT RAISE(ABORT, 'insert on table gpkg_metadata_reference violates constraint: column name must be NULL when reference_scope is "geopackage","table" or "row"') WHERE (NEW.reference_scope IN ('geopackage','table','row') AND NEW.column_name IS NOT NULL); SELECT RAISE(ABORT, 'insert on table gpkg_metadata_reference violates constraint: column name must be defined for the specified table when reference_scope is "column" or "row/col"') WHERE (NEW.reference_scope IN ('column','row/col') AND NOT NEW.table_name IN (SELECT name FROM SQLITE_MASTER WHERE type = 'table' AND name = NEW.table_name AND sql LIKE ('%' || NEW.column_name || '%'))); END""",
    """CREATE TRIGGER IF NOT EXISTS 'gpkg_metadata_reference_column_name_update' BEFORE UPDATE OF column_name ON 'gpkg_metadata_reference' FOR EACH ROW BEGIN SELECT RAISE(ABORT, 'update on table gpkg_metadata_reference violates constraint: column name must be NULL when reference_scope is "geopackage","table" or "row"') WHERE (NEW.reference_scope IN ('geopackage','table','row') AND NEW.column_name IS NOT NULL); SELECT RAISE(ABORT, 'update on table gpkg_metadata_reference violates constraint: column name must be defined for the specified table when reference_scope is "column" or "row/col"') WHERE (NEW.reference_scope IN ('column','row/col') AND NOT NEW.table_name IN (SELECT name FROM SQLITE_MASTER WHERE type = 'table' AND name = NEW.table_name AND sql LIKE ('%' || NEW.column_name || '%'))); END""",
    """CREATE TRIGGER IF NOT EXISTS 'gpkg_metadata_reference_reference_scope_insert' BEFORE INSERT ON 'gpkg_metadata_reference' FOR EACH ROW BEGIN SELECT RAISE(ABORT, 'insert on table gpkg_metadata_reference violates constraint: reference_scope must be one of "geopackage", "table","column","row", "row/col"') WHERE NOT NEW.reference_scope IN ('geopackage','table','column','row','row/col');END""",
    """CREATE TRIGGER IF NOT EXISTS 'gpkg_metadata_reference_reference_scope_update' BEFORE UPDATE OF 'reference_scope' ON 'gpkg_metadata_reference' FOR EACH ROW BEGIN SELECT RAISE(ABORT, 'update on table gpkg_metadata_reference violates constraint: reference_scope must be one of "geopackage", "table","column","row", "row/col"') WHERE NOT NEW.reference_scope IN ('geopackage','table','column','row','row/col');END""",
    """CREATE TRIGGER IF NOT EXISTS 'gpkg_metadata_reference_timestamp_insert' BEFORE INSERT ON 'gpkg_metadata_reference' FOR EACH ROW BEGIN SELECT RAISE(ABORT, 'insert on table gpkg_metadata_reference violates constraint: timestamp must be a valid time in ISO 8601 "yyyy-mm-ddThh-mm-ss.cccZ" form') WHERE NOT (NEW.timestamp GLOB '[1-2][0-9][0-9][0-9]-[0-1][0-9]-[1-3][0-9]T[0-2][0-9]:[0-5][0-9]:[0-5][0-9].[0-9][0-9][0-9]Z' AND strftime('%s',NEW.timestamp) NOT NULL);END""",
    """CREATE TRIGGER IF NOT EXISTS 'gpkg_metadata_reference_timestamp_update' BEFORE UPDATE OF 'timestamp' ON 'gpkg_metadata_reference' FOR EACH ROW BEGIN SELECT RAISE(ABORT, 'update on table gpkg_metadata_reference violates constraint: timestamp must be a valid time in ISO 8601 "yyyy-mm-ddThh-mm-ss.cccZ" form') WHERE NOT (NEW.timestamp GLOB '[1-2][0-9][0-9][0-9]-[0-1][0-9]-[1-3][0-9]T[0-2][0-9]:[0-5][0-9]:[0-5][0-9].[0-9][0-9][0-9]Z' AND strftime('%s',NEW.timestamp) NOT NULL);END""",
    """CREATE TRIGGER IF NOT EXISTS 'gpkg_tile_matrix_matrix_height_insert' BEFORE INSERT ON 'gpkg_tile_matrix' FOR EACH ROW BEGIN SELECT RAISE(ABORT, 'insert on table ''gpkg_tile_matrix'' violates constraint: matrix_height cannot be less than 1') WHERE (NEW.matrix_height < 1);END""",
    """CREATE TRIGGER IF NOT EXISTS 'gpkg_tile_matrix_matrix_height_update' BEFORE UPDATE OF matrix_height ON 'gpkg_tile_matrix' FOR EACH ROW BEGIN SELECT RAISE(ABORT, 'update on table ''gpkg_tile_matrix'' violates constraint: matrix_height cannot be less than 1') WHERE (NEW.matrix_height < 1);END""",
    """CREATE TRIGGER IF NOT EXISTS 'gpkg_tile_matrix_matrix_width_insert' BEFORE INSERT ON 'gpkg_tile_matrix' FOR EACH ROW BEGIN SELECT RAISE(ABORT, 'insert on table ''gpkg_tile_matrix'' violates constraint: matrix_width cannot be less than 1') WHERE (NEW.matrix_width < 1);END""",
    """CREATE TRIGGER IF NOT EXISTS 'gpkg_tile_matrix_matrix_width_update' BEFORE UPDATE OF matrix_width ON 'gpkg_tile_matrix' FOR EACH ROW BEGIN SELECT RAISE(ABORT, 'update on table ''gpkg_tile_matrix'' violates constraint: matrix_width cannot be less than 1') WHERE (NEW.matrix_width < 1);END""",
    """CREATE TRIGGER IF NOT EXISTS 'gpkg_tile_matrix_pixel_x_size_insert' BEFORE INSERT ON 'gpkg_tile_matrix' FOR EACH ROW BEGIN SELECT RAISE(ABORT, 'insert on table ''gpkg_tile_matrix'' violates constraint: pixel_x_size must be greater than 0') WHERE NOT (NEW.pixel_x_size > 0);END""",
    """CREATE TRIGGER IF NOT EXISTS 'gpkg_tile_matrix_pixel_x_size_update' BEFORE UPDATE OF pixel_x_size ON 'gpkg_tile_matrix' FOR EACH ROW BEGIN SELECT RAISE(ABORT, 'update on table ''gpkg_tile_matrix'' violates constraint: pixel_x_size must be greater than 0') WHERE NOT (NEW.pixel_x_size > 0);END""",
    """CREATE TRIGGER IF NOT EXISTS 'gpkg_tile_matrix_pixel_y_size_insert' BEFORE INSERT ON 'gpkg_tile_matrix' FOR EACH ROW BEGIN SELECT RAISE(ABORT, 'insert on table ''gpkg_tile_matrix'' violates constraint: pixel_y_size must be greater than 0') WHERE NOT (NEW.pixel_y_size > 0);END""",
    """CREATE TRIGGER IF NOT EXISTS 'gpkg_tile_matrix_pixel_y_size_update' BEFORE UPDATE OF pixel_y_size ON 'gpkg_tile_matrix' FOR EACH ROW BEGIN SELECT RAISE(ABORT, 'update on table ''gpkg_tile_matrix'' violates constraint: pixel_y_size must be greater than 0') WHERE NOT (NEW.pixel_y_size > 0);END""",
    """CREATE TRIGGER IF NOT EXISTS 'gpkg_tile_matrix_zoom_level_insert' BEFORE INSERT ON 'gpkg_tile_matrix' FOR EACH ROW BEGIN SELECT RAISE(ABORT, 'insert on table ''gpkg_tile_matrix'' violates constraint: zoom_level cannot be less than 0') WHERE (NEW.zoom_level < 0);END""",
    """CREATE TRIGGER IF NOT EXISTS 'gpkg_tile_matrix_zoom_level_update' BEFORE UPDATE OF zoom_level ON 'gpkg_tile_matrix' FOR EACH ROW BEGIN SELECT RAISE(ABORT, 'update on table ''gpkg_tile_matrix'' violates constraint: zoom_level cannot be less than 0') WHERE (NEW.zoom_level < 0);END"""
]
#----------------------------------------------------------------------
def _insert_values(con, tbl, fields, values):
    """inserts multiple values into a table"""
    q = ["?"] * len(fields)
    q = ",".join(q)
    sql = '''INSERT INTO {table} ({fields})
                 VALUES({q})'''.format(table=tbl,
                                       fields=",".join(fields),
                                       q=q)
    inserts = []
    cur = con.cursor()
    for val in values:
        if not isinstance(val, list):
            val = [val]
        cur.execute(sql, val)
        inserts.append(cur.lastrowid)
    con.commit()
    return inserts
#--------------------------------------------------------------------------
def _copy_table_schema(con, table_source, table_name_dest):
    """copies a table's schema to a new table"""
    from ._geopackage import SpatialTable, Table
    sql = """CREATE TABLE {dest} AS SELECT * FROM {source} WHERE 0""".format(source=table_source,
                                                                             dest=table_name_dest)
    #_insert_values(con=con, tbl="gpkg_contents",
                   #fields=['table_name', 'data_type', 'identifier', 'srs_id'],
                   #values=[[name, 'features', name, wkid]])

    #_insert_values(con=con, tbl="gpkg_geometry_columns",
                   #fields=['table_name', 'column_name', 'geometry_type_name', 'srs_id', 'z', 'm'],
                   #values=[[name, 'Shape', _geom_lu[geometry.lower()], wkid, int(has_z),int(has_m)]])
    try:
        con.execute(sql)
        con.commit()
    except Exception as e:
        raise Exception(e)


#----------------------------------------------------------------------
def _create_gpkg(name, path=None, overwrite=False):
    """
    Creates an empty geopackage
    """
    if path is None:
        path = tempfile.gettempdir()
    elif os.path.isdir(path) == False and \
         len(path) > 0:
        os.makedirs(path, exist_ok=True)
    if name.endswith(".gpkg") == False:
        name += ".gpkg"
    fp = os.path.join(path, name)
    if overwrite and os.path.isfile(fp):
        os.remove(fp)
    con = sqlite3.connect(database=fp)

    for sql in _create_tables_sql:
        con.execute(sql)
    for trg_sql in _initial_triggers_sql:
        con.execute(trg_sql)
    sql = """SELECT * from gpkg_extensions"""
    cur = con.execute(sql)
    result = cur.fetchall()
    con.commit()
    del cur
    if len(result) < 2:
        _insert_values(con=con,
                       tbl="gpkg_extensions",
                       fields=["extension_name", "definition", "scope"],
                       values=[
                           ["gpkg_metadata", 'F.8 Metadata', 'read-write'],
                           ["gpkg_schema", 'F.9 Schema', 'read-write']
                       ]
                       )
        _insert_values(con=con,
                      tbl="gpkg_spatial_ref_sys",
                      fields=["srs_name", "srs_id", "organization", "organization_coordsys_id", "definition", "description"],
                      values=[
                          ["GCS_WGS_1984", 4326, "EPSG", "4326", """GEOGCS["GCS_WGS_1984",DATUM["D_WGS_1984",SPHEROID["WGS_1984",6378137.0,298.257223563]],PRIMEM["Greenwich",0.0],UNIT["Degree",0.0174532925199433]]""", "WGS 1984"],
                          ["Undefined Cartesian", -1, "NONE", "-1", "undefined", None],
                          ["Undefined Geographic", 0, "NONE", "0", "undefined", None]
                      ])
    con.commit()
    con.close()
    del con
    return fp
#--------------------------------------------------------------------------
def _create_table(con, name, fields=None):
    """creates an attribute table"""
    txts = ["OBJECTID INTEGER primary key autoincrement not null"]
    if fields is None:
        fields = {}

    for k,v in fields.items():
        row = _field_lookup[v.lower()]
        if row[0].lower() != "text":
            fld = "{field} {dtype} {st}".format(field=k, dtype=row[0], st=row[1])
            txts.append(fld.format(field=k))
        else:
            fld = "{field} {dtype}".format(field=k, dtype=row[0])
            txts.append(fld)
        del k,v
    try:
        sql = """CREATE TABLE IF NOT EXISTS {tbl} ({fields})""".format(tbl=name, fields=",".join(txts))
        con.execute(sql)
        con.commit()
        _insert_values(con=con, tbl="gpkg_contents",
                       fields=['table_name', 'data_type', 'identifier'],
                       values=[[name, 'attributes', name]])
        return True
    except:
        return False
#----------------------------------------------------------------------
def _create_feature_class(con,
                          name,
                          geometry="point",
                          wkid=4326,
                          fields=None,
                          has_z=False,
                          has_m=False):
    """
    Creates a spatial table

    ==============     ====================================================
    **Argument**       **Description**
    --------------     ----------------------------------------------------
    name               Required String.  The name of the table.
    --------------     ----------------------------------------------------
    geometry           Optional String. The type of geometry to create.
                       This can be point, line, polygon or multipoint.
    --------------     ----------------------------------------------------
    wkid               Optional Integer.  The spatial reference code.
    ==============     ====================================================

    """
    _geom_lu = {
        "point" : "POINT",
        "multipoint" : "MULTIPOINT",
        "polygon" : "MULTIPOLYGON",
        "polyline" : "POLYLINE",
        "line" : "POLYLINE"
    }
    txts = ["OBJECTID INTEGER primary key autoincrement not null",
            "'Shape' {gt}".format(gt=_geom_lu[geometry.lower()])]
    if fields is None:
        fields = {}

    for k,v in fields.items():
        row = _field_lookup[v.lower()]
        if row[0].lower() != "text":
            fld = "{field} {dtype} {st}".format(field=k, dtype=row[0], st=row[1])
            txts.append(fld.format(field=k))
        else:
            fld = "{field} {dtype}".format(field=k, dtype=row[0])
            txts.append(fld)
        del k,v
    sql = """CREATE TABLE IF NOT EXISTS {name} ({fields})""".format(name=name,
                                                                    fields=",".join(txts))
    try:
        con.execute(sql)
        con.commit()
    except Exception as e:
        raise Exception(e)

    _insert_values(con=con, tbl="gpkg_contents",
                   fields=['table_name', 'data_type', 'identifier', 'srs_id'],
                   values=[[name, 'features', name, wkid]])

    _insert_values(con=con, tbl="gpkg_geometry_columns",
                   fields=['table_name', 'column_name', 'geometry_type_name', 'srs_id', 'z', 'm'],
                   values=[[name, 'Shape', _geom_lu[geometry.lower()], wkid, int(has_z),int(has_m)]])

    sql = """select srs_id from gpkg_spatial_ref_sys where srs_id = {wkid};""".format(wkid=wkid)
    sr_check = len(con.execute(sql).fetchall())
    if sr_check == 0:
        from ._coord import lookup_coordinate_system
        res = lookup_coordinate_system(wkid=wkid)
        if len(res) > 0:
            res = res[0]
        _insert_values(con=con, tbl="gpkg_spatial_ref_sys",
                       fields=['srs_name', 'srs_id',
                               'organization', 'description',
                               'organization_coordsys_id', 'definition'],
                       values=[[res['NAME'], res['WKID'], "ESRI",
                                res['NAME'], res['WKID'], res['WKT']]])
    return True



"""
Defined functions for r-tree index as specified in
https://github.com/opengeospatial/geopackage/blob/master/spec/annexes/extension_spatialindex.adoc
"""
from ._wkb import loads as _loads
from ._wkb import dumps as _dumps
try:
    from geomet.wkb import loads as _geomet_loads
    from geomet.wkb import dumps as _geomet_dumps
except:
    pass
#--------------------------------------------------------------------------
def _strip_header(g):
    """removes the GP header"""
    if g[:2] == b'GB':
        return g[8:]
    return g
def _flatten_list(coords):
    """moves all x,y pairs into two lists"""
    minx, maxx = None, None
    miny, maxy = None, None
    for prt in coords:
        for xy in prt:
            if miny is None and maxy is None:
                miny = xy[1]
                maxy = xy[1]
            if minx is None and maxx is None:
                minx = xy[0]
                maxx = xy[0]

    return [],[]
def _extent(g, t='xmin'):
    """gets the extent value from the geometry"""
    geom = _strip_header(geom)
    j = _dumps(geom)
    if t == 'xmin':
        if 'x' in j:
            return j['x']
        elif 'rings' in j:

            return

    elif 'y' in j:
        return j['y']
#--------------------------------------------------------------------------
def ST_IsEmpty(geom):
    """
    returns 1 if geometry value is empoty 0 if not empty, NULL if geometry is NULL
    """
    if geom is None:
        return None
    geom = _strip_header(geom)
    if geom == b'0x000000000000f87f':
        return 1
    return 0
#--------------------------------------------------------------------------
def ST_MinX(geom):
    """
    Returns the minimum X value of the bounding envelope of a geometry
    """
    geom = _strip_header(geom)
    j = _dumps(geom)

    return
#--------------------------------------------------------------------------
def ST_MinY(geom):
    """
    Returns the minimum Y value of the bounding envelope of a geometry
    """
    geom = _strip_header(geom)
    return
#--------------------------------------------------------------------------
def ST_MaxX(geom):
    """
    Returns the maximum X value of the bounding envelope of a geometry
    """
    geom = _strip_header(geom)
    return
#--------------------------------------------------------------------------
def ST_MaxY(geom):
    """
    Returns the maximum Y value of the bounding envelope of a geometry
    """
    geom = _strip_header(geom)
    return


import os
import sys
import json
import sqlite3

import tempfile
from io import BytesIO, StringIO
from sqlite3 import Binary as sBinary
from collections import OrderedDict, MutableMapping

try:
    import geomet
    from geomet import wkt as geometwkt
    from geomet import wkb as geometwkb
    _HASGEOMET = True
except ImportError:
    _HASGEOMET = False

from ._gpkg import _create_feature_class, _create_gpkg, _create_table, _insert_values
from ._wkb import loads, dumps
# ----------------------------------------------------------------------
def _handle_wkb(wkb):
    """handles the insert convertion for the custom geometry types"""
    if isinstance(wkb, bytes):
        wkb = bytearray(wkb)
    return wkb
# ----------------------------------------------------------------------
def _adapt_wkb(wkb):
    """ensures the wkb values are bytes, not bytearrays"""
    return wkb
########################################################################
class GeoPackage(object):
    """
    A single instance of a GeoPackage file.
    """
    _con = None
    _dir = None
    _path = None
    _db_name = None

    #----------------------------------------------------------------------
    def __init__(self, path, overwrite=False):
        """Constructor"""

        self._dir = os.path.dirname(path)
        self._db_name = os.path.basename(path)

        if self._db_name.lower().endswith('.gpkg') == False:
            self._db_name += ".gpkg"
        self._path = os.path.join(self._dir, self._db_name)
        if os.path.isfile(self._path) and \
           overwrite:
            os.remove(self._path)
        self._path = _create_gpkg(name=self._db_name,
                                  path=self._dir,
                                  overwrite=overwrite)
        self._con = sqlite3.connect(self._path,
                                    detect_types=sqlite3.PARSE_DECLTYPES)
        # register custom dtypes
        sqlite3.register_adapter(bytearray, _adapt_wkb)
        for g in ["POINT", "LINESTRING", "POLYGON",
                  "MULTIPOINT", "MULTILINESTRING", "MULTIPOLYGON"]:
            sqlite3.register_converter(g, _handle_wkb)
    #----------------------------------------------------------------------
    def _setup(self):
        """sets up the registration for the GeoPackage"""
        if self._path != ":memory:":
            self._con.close()
            self._con = None
            self._con = sqlite3.connect(self._path,
                                        detect_types=sqlite3.PARSE_DECLTYPES)
        # register custom dtypes
        sqlite3.register_adapter(bytearray, _adapt_wkb)
        for g in ["POINT", "LINESTRING", "POLYGON",
                      "MULTIPOINT", "MULTILINESTRING", "MULTIPOLYGON"]:
            sqlite3.register_converter(g, _handle_wkb)
    #----------------------------------------------------------------------
    def __len__(self):
        """returns the number of registered tables"""
        try:
            sql = """SELECT count(*) FROM gpkg_contents"""
            cur = self._con.execute(sql)
            return cur.fetchone()[0]
        except:
            return 0
    #----------------------------------------------------------------------
    def __enter__(self):
        if self._con is None:
            self._con = sqlite3.connect(self._path)
        return self
    #----------------------------------------------------------------------
    def __exit__(self, type, value, traceback):
        self._con.commit()
        self._con.close()
    #----------------------------------------------------------------------
    def exists(self, name):
        """
        Returns boolean if the table exists

        :returns: boolean

        """
        sql = "SELECT table_name, data_type FROM gpkg_contents"
        cur = self._con.execute(sql)
        for tbl in cur:
            if tbl[0].lower() == name.lower():
                return True
        return False
    #----------------------------------------------------------------------
    def get(self, name):
        """
        Returns a table if it exists in the geopackage.

        ===============     ===============================================
        **Arguements**      **Description**
        ---------------     -----------------------------------------------
        name                Optional String. The name of the table or
                            feature class.
        ===============     ===============================================

        :returns: Table/SpatialTable

        """
        sql = "SELECT table_name, data_type FROM gpkg_contents where table_name = ?"
        cur = self._con.execute(sql, [name])
        for tbl in cur:
            if tbl[1] == 'attributes':
                return Table(tbl[0], self._con)
            else:
                return SpatialTable(tbl[0], self._con)
    #----------------------------------------------------------------------
    @property
    def tables(self):
        """
        Gets a list of registered table names with the geopackage

        :returns: iterator

        """
        sql = "SELECT table_name, data_type FROM gpkg_contents"
        cur = self._con.execute(sql)
        for tbl in cur:
            if tbl[1] == 'attributes':
                yield Table(tbl[0], self._con)
            else:
                yield SpatialTable(tbl[0], self._con)
    #----------------------------------------------------------------------
    def enable(self):
        """
        enables the sqlite database to be a geopackage

        :returns: Boolean

        """
        try:
            self._setup()
        except:
            return False
        return True
    #----------------------------------------------------------------------
    def create(self,
               name,
               fields=None,
               wkid=None,
               geometry_type=None,
               overwrite=True):
        """
        The `create` method generates a new table or feature class in the
        geopackage.

        ===============     ===============================================
        **Arguements**      **Description**
        ---------------     -----------------------------------------------
        name                Optional String. The name of the table or
                            feature class.
        ---------------     -----------------------------------------------
        fields              Optional dict.  The columns to add to a table.
                            An OBJECTID field is always created for any table.

                            Allowed Fields:

                            + TEXT -Any string of characters.
                            + FLOAT - Fractional numbers between -3.4E38 and 1.2E38.
                            + DOUBLE - Fractional numbers between -2.2E308 and 1.8E308.
                            + SHORT - Whole numbers between -32,768 and 32,767.
                            + LONG - Whole numbers between -2,147,483,648 and 2,147,483,647.
                            + DATE -Date and/or time.
                            + BLOB -Long sequence of binary numbers.
                            + GUID -Globally unique identifier.

        ---------------     -----------------------------------------------
        wkid                Optional Int. The SRS code for the feature class.
        ---------------     -----------------------------------------------
        geometry_type       Optional String. If given the output will be a
                            SpatialTable instead of a Table. Allowed values
                            are: point, line, polygon, and multipoint.
        ---------------     -----------------------------------------------
        overwrite           Optional Boolean. If True, the geopackage will
                            attempt to erase the table and recreate it with
                            the new schema.  All records from the old table
                            will be lost.
        ===============     ===============================================

        :returns: Table/SpatialTable
        """
        if overwrite:
            sql_drop = """DROP TABLE IF EXISTS %s""" % name
            sql_delete_row = """DELETE FROM gpkg_contents where table_name = '{tbl}'""".format(tbl=name)
            sql_geom_col = """DELETE FROM gpkg_geometry_columns where table_name = '{tbl}'""".format(tbl=name)
            self._con.execute(sql_drop)
            self._con.execute(sql_delete_row)
            self._con.execute(sql_geom_col)
            self._con.commit()
        elif self.exists(name) and \
             overwrite == False:
            raise ValueError("Table %s exists. Please pick a different table name" % name)
        if geometry_type:
            iftrue = _create_feature_class(con=self._con,
                                           name=name,
                                           wkid=wkid,
                                           fields=fields,
                                           geometry=geometry_type)
            if iftrue:
                return SpatialTable(table=name,
                                    con=self._con)

        else:
            iftrue = _create_table(con=self._con,
                                 name=name,
                                 fields=fields)
            if iftrue:
                return Table(table=name, con=self._con)

        return
########################################################################
class _Row(MutableMapping, OrderedDict):
    """
    A Single Row Entry. This class is created by the `Table` class.

    ** It should not be created by a user. **
    """
    _con = None
    _values = None
    _table_name = None
    _dict = None
    _keys = None
    #----------------------------------------------------------------------
    def __init__(self, values, table_name=None, con=None, header=None):
        """Constructor"""
        self._table_name = table_name
        self._con = con
        self._values = values
        self._header = header
    #----------------------------------------------------------------------
    def __str__(self):
        return "<Row OBJECTID=%s>" % self['OBJECTID']
    #----------------------------------------------------------------------
    def __repr__(self):
        return self.__str__()
    #----------------------------------------------------------------------
    def __setattr__(self, name, value):
        if name in {'_values','_dict', '_table_name', '_con', '_keys', '_header'}:
            super().__setattr__(name, value)
        elif name.lower() == 'shape':
            self._values[name] = value
            self._update()
        elif name.lower() != 'objectid' and \
             name in self.keys():
            self._values[name] = value
            self._update()

        elif name.lower() == 'objectid':
            raise ValueError("OBJECTID values cannot be updated.")
        else:
            raise ValueError("The field: {field} does not exist.".format(field=name))
    #----------------------------------------------------------------------
    def __getattr__(self, name):
        if name in self._values:
            return self._values[name]
        return
    #----------------------------------------------------------------------
    def __getitem__(self, name):
        return self.__getattr__(name)
    #----------------------------------------------------------------------
    def __setitem__(self, name, value):
        self.__setattr__(name, value)
    #----------------------------------------------------------------------
    def keys(self):
        """returns the column names in the dataset"""
        return list(self._values.keys())
    #----------------------------------------------------------------------
    @property
    def fields(self):
        """returns the field names in the dataset"""
        return self.keys()
    #----------------------------------------------------------------------
    def as_dict(self):
        """returns the row as a dictionary"""
        return dict(zip(self.keys(), self.values()))
    #----------------------------------------------------------------------
    def values(self):
        """returns the row values"""
        return list(self._values.values())
    #----------------------------------------------------------------------
    def _update(self):
        """updates the current row"""
        txts = []
        values = []
        for k,v in self._values.items():
            if k.lower() != "objectid" and \
               k.lower() != 'shape':
                txts.append("%s=?" % k)
                values.append(v)
            elif k.lower() == 'shape':
                if isinstance(v, dict) and "coordinates" not in v:
                    v = self._header + dumps(v, False)
                elif isinstance(v, dict) and "coordinates" in v:
                    v = self._gpheader + geometwkb.dumps(obj=v)
                elif isinstance(v, str):
                    gj = geometwkt.loads(v)
                    v = self._gpheader + geometwkb.dumps(obj=gj)
                elif isinstance(v, (bytes, bytearray)):
                    if isinstance(v, (bytearray)):
                        v = bytes(v)
                    if len(v) > 2 and \
                       v[:2] != b'GB':
                        v = self._gpheader + v
                elif v is None:
                    v = self._gpheader + b'0x000000000000f87f'
                else:
                    raise ValueError(("Shape column must be Esri JSON dictionary, "
                                      "WKT, GeoJSON dictionary, or WKB (bytes)"))
                txts.append("%s=?" % k)
                values.append(v)
        sql = '''UPDATE {table} SET {values} WHERE OBJECTID={oid}'''.format(
            table=self._table_name,
            values=",".join(txts),
            oid=self._values['OBJECTID'])
        cur = self._con.execute(sql, values)
        self._con.commit()
        del sql
        del values, txts
        return True
    #----------------------------------------------------------------------
    def delete(self):
        """
        Deletes the current row

        :returns: Boolean
        """
        try:
            cur = self._con.execute(
                '''DELETE FROM {tbl} WHERE OBJECTID=?'''.format(tbl=self._table_name),
                [self['OBJECTID']])
            return True
        except:
            return False
########################################################################
class Table(object):
    """
    A Table object is a attribute only set of data. No spatial data is associated with
    this information.


    ===============     ===============================================
    **Arguements**      **Description**
    ---------------     -----------------------------------------------
    table               Requred String. The name of the table.
    ---------------     -----------------------------------------------
    con                 Required sqlite3.Connection.  The active connection
                        to the geopackage.
    ===============     ===============================================

    """

    _con = None
    _table_name = None
    _create_sql = None
    _fields = None
    #----------------------------------------------------------------------
    def __init__(self, table, con):
        """Constructor"""
        self._con = con
        self._table_name = table
    #----------------------------------------------------------------------
    @property
    def dtype(self):
        """returns the table type"""
        return "attribute"
    #----------------------------------------------------------------------
    def __str__(self):
        return "<Attribute Table: {table}>".format(table=self._table_name)
    #----------------------------------------------------------------------
    def __repr__(self):
        return self.__str__()
    #----------------------------------------------------------------------
    def __iter__(self):
        for row in self.rows():
            yield row
    #----------------------------------------------------------------------
    @property
    def fields(self):
        """
        returns the field information for a table

        :returns: Dictionary

        """
        if self._fields is None:
            sql = """PRAGMA table_info({tbl});""".format(tbl=self._table_name)
            rows = self._con.execute(sql).fetchall()
            self._fields = {row[1]: row[2] for row in rows}
        return self._fields
    #----------------------------------------------------------------------
    def add_field(self, name, data_type):
        """

        Adds a new column to the table.

        ===============     ===============================================
        **Arguements**      **Description**
        ---------------     -----------------------------------------------
        name                Required String. The name of the field.
        ---------------     -----------------------------------------------
        data_types          Required String.  The type of column to add.

                            Allowed Data Types:

                            + TEXT -Any string of characters.
                            + FLOAT - Fractional numbers between -3.4E38 and 1.2E38.
                            + DOUBLE - Fractional numbers between -2.2E308 and 1.8E308.
                            + SHORT - Whole numbers between -32,768 and 32,767.
                            + LONG - Whole numbers between -2,147,483,648 and 2,147,483,647.
                            + DATE -Date and/or time.
                            + BLOB -Long sequence of binary numbers.
                            + GUID -Globally unique identifier.
        ===============     ===============================================

        :returns: Boolean

        """
        _field_lookup = {
            "text" : ["TEXT", "check((typeof({field}) = 'text' or typeof({field}) = 'null') and not length({field}) > {l})"],
            "float" : ["DOUBLE", """check(typeof({field}) = 'real' or typeof({field}) = 'null')"""],
            "double" : ["DOUBLE", "check(typeof({field}) = 'real' or typeof({field}) = 'null')"],
            "short" : ["SMALLINT", "check((typeof({field}) = 'integer' or typeof({field}) = 'null') and {field} >= -32768 and {field} <= 32767)"],
            "long" : ["MEDIUMINT", "check((typeof({field}) = 'integer' or typeof({field}) = 'null') and {field} >= -2147483648 and {field} <= 2147483647)"],
            "integer" : ["MEDIUMINT", "check((typeof({field}) = 'integer' or typeof({field}) = 'null') and {field} >= -2147483648 and {field} <= 2147483647)"],
            "date" : ["DATETIME", "check((typeof({field}) = 'text' or typeof({field}) = 'null') and strftime('%Y-%m-%dT%H:%M:%fZ',{field}))"],
            "blob" : ["BLOB", "check(typeof({field}) = 'blob' or typeof({field}) = 'null')"],
            "guid" : ["TEXT(38)", "check((typeof({field}) = 'text' or typeof({field}) = 'null') and not length({field}) > 38)"]
        }
        try:
            row = _field_lookup[data_type.lower()]
            if row[0].lower() != "text":
                fld = ("{field} {dtype} {st}"
                       .format(field=name, dtype=row[0], st=row[1])
                       .format(field=name))
            else:
                fld = "{field} {dtype}".format(field=name, dtype=row[0])
            sql = """ALTER TABLE {table} ADD COLUMN {dtype};""".format(table=self._table_name,
                                                                          field=name,
                                                                          dtype=fld)
            self._con.execute(sql)
            self._con.commit()
            self._fields = None
            return True
        except:
            return False
    #----------------------------------------------------------------------
    def delete_field(self, name):
        """
        Drops a Field from a Table

        ===============     ===============================================
        **Arguements**      **Description**
        ---------------     -----------------------------------------------
        name                Required String. The name of the field to remove.
        ===============     ===============================================

        :returns: boolean
        """
        fields = ",".join([fld for fld in self.fields if fld.lower() != name.lower()])
        sql = """
        CREATE TABLE temp_bkup AS SELECT {fields} FROM {table};
        DROP TABLE {table};
        ALTER TABLE temp_bkup RENAME TO {table};
        """.format(table=self._table_name, fields=fields)
        self._con.executescript(sql)
        self._con.commit()
        self._fields = None
        return True
    #----------------------------------------------------------------------
    def rows(self, where=None, fields="*"):
        """
        Search/update cursor like iterator

        ===============     ===============================================
        **Arguements**      **Description**
        ---------------     -----------------------------------------------
        where               Optional String. Optional Sql where clause.
        ---------------     -----------------------------------------------
        fields              Optional List. The default is all fields (*).
                            A list of fields can be provided to limit the
                            data that is returned.
        ===============     ===============================================

        :returns: _Row object
        """
        if isinstance(fields, (list, tuple)):
            if "OBJECTID" not in fields:
                fields.append("OBJECTID")
            fields = ",".join(fields)
        if where is None:
            query = """SELECT {fields} from {tbl} """.format(tbl=self._table_name,
                                                             fields=fields)
        else:
            query = """SELECT {fields} from {tbl} WHERE {where}""".format(tbl=self._table_name,
                                                                          fields=fields,
                                                                          where=where)
        cursor = self._con.cursor()
        c = cursor.execute(query)
        columns = [d[0] for d in c.description]
        for row in c:
            yield _Row(values=dict(zip(columns, row)),
                       table_name=self._table_name,
                       con=self._con)
    #----------------------------------------------------------------------
    def insert(self, row):
        """
        Inserts a new row via dictionary

        ===============     ===============================================
        **Arguements**      **Description**
        ---------------     -----------------------------------------------
        row                 Required Dictionary.  Insert a new row via a
                            dictionary. The key/value pair must match up to
                            the field names in the table.
        ==============      ===============================================

        :returns: Boolean

        """
        values = None
        if isinstance(row, dict):
            keys = row.keys()
            values = [list(row.values())]
        elif isinstance(row, (list, tuple)):
            keys = row[0].keys()
            values = [list(r.values()) for r in row]
        q = ["?"] * len(keys)
        q = ",".join(q)
        sql = '''INSERT INTO {table} ({fields})
                     VALUES({q})'''.format(table=self._table_name,
                                           fields=",".join(keys),
                                           q=q)
        inserts = []
        cur = self._con.cursor()
        cur.execute(sql, list(values[0]))
        self._con.commit()
        return True
    #----------------------------------------------------------------------
    def to_pandas(self, where=None, fields="*", ftype=None):
        """
        Exports a table to a Pandas' DataFrame.

        ===============     ===============================================
        **Arguements**      **Description**
        ---------------     -----------------------------------------------
        where               Optional String. Optional Sql where clause.
        ---------------     -----------------------------------------------
        fields              Optional List. The default is all fields (*).
                            A list of fields can be provided to limit the
                            data that is returned.
        ---------------     -----------------------------------------------
        ftype               Optional String. This value can be dataframe
                            format type.  The value can be None or esri.

                                + None - means the dataframe will be a raw view of the table.
                                + esri - means the dataframe will be a spatially enable dataframe. (Requires Python API for ArcGIS)

        ===============     ===============================================

        :returns: pd.DataFrame

        """
        import pandas as pd
        if isinstance(fields, (list, tuple)):
            if "OBJECTID" not in fields:
                fields.append("OBJECTID")
            fields = ",".join(fields)
        if where is None:
            query = """SELECT {fields} from {tbl} """.format(tbl=self._table_name,
                                                             fields=fields)
        else:
            query = """SELECT {fields} from {tbl} WHERE {where}""".format(tbl=self._table_name,
                                                                          fields=fields,
                                                                          where=where)
        if ftype is None:
            return pd.read_sql_query(query,
                                     self._con)
        elif str(ftype).lower() == 'esri':
            try:
                from arcgis.geometry import Geometry
                from arcgis.features import GeoAccessor, GeoSeriesAccessor
                df = pd.read_sql_query(query,
                                       self._con)
                fields = list(self.fields.keys())
                lower_fields = [str(fld).lower() for fld in fields]
                if "shape" in lower_fields:
                    idx = lower_fields.index("shape")
                    SHAPE = fields[idx]
                    df[SHAPE] = df[SHAPE].apply(lambda x: Geometry(x[8:]))
                    df.spatial.set_geometry(SHAPE)
                    try:
                        df.spatial.project(self.wkid)
                    except:
                        print('nope')
                return df
            except ImportError:
                raise Exception("The Python API for ArcGIS is required to import using ftype `esri`")
            except Exception as e:
                raise Exception(e)
########################################################################
class SpatialTable(Table):
    """
    Represents a Feature Class inside a GeoPackage

    """
    _con = None
    _wkid = None
    _gtype = None
    _fields = None
    _table_name = None
    _create_sql = None
    _gp_header = None
    #----------------------------------------------------------------------
    def __init__(self, table, con):
        """Constructor"""
        self._table_name = table
        self._con = con
        self._refresh()
    #----------------------------------------------------------------------
    def _refresh(self):
        """internal method that refreshes the table information"""
        self._sd_lu = """SELECT * from gpkg_geometry_columns where table_name = '%s'""" % self._table_name
        cur = self._con.execute(self._sd_lu)
        self._gp_header = None

        for row in cur:
            self._gtype = row[2]
            self._wkid = row[3]
            break
        del cur
    #----------------------------------------------------------------------
    def __str__(self):
        return "<Spatial Table: {table}, {gt}>".format(table=self._table_name,
                                                       gt=self._gtype)
    #----------------------------------------------------------------------
    def __repr__(self):
        return self.__str__()
    #----------------------------------------------------------------------
    @property
    def geometry_type(self):
        if self._gtype is None:
            self._refresh()
        return self._gtype
    #----------------------------------------------------------------------
    @property
    def dtype(self):
        """
        Returns the table type

        :returns: String

        """
        return "spatial"
    #----------------------------------------------------------------------
    @property
    def wkid(self):
        """
        Returns the Spatial Table's WKID/SRS ID

        :returns: Integer
        """
        if self._wkid is None:
            self._refresh()
        return self._wkid
    #----------------------------------------------------------------------
    def rows(self, where=None, fields="*"):
        """
        Search/update cursor like iterator

        ===============     ===============================================
        **Arguements**      **Description**
        ---------------     -----------------------------------------------
        where               Optional String. Optional Sql where clause.
        ---------------     -----------------------------------------------
        fields              Optional List. The default is all fields (*).
                            A list of fields can be provided to limit the
                            data that is returned.
        ===============     ===============================================

        :returns: _Row object
        """
        if isinstance(fields, (list, tuple)):
            if "OBJECTID" not in fields:
                fields.append("OBJECTID")
            fields = ",".join(fields)
        if where is None:
            query = """SELECT {fields} from {tbl} """.format(tbl=self._table_name,
                                                             fields=fields)
        else:
            query = """SELECT {fields} from {tbl} WHERE {where}""".format(tbl=self._table_name,
                                                                          fields=fields,
                                                                          where=where)
        cursor = self._con.cursor()
        c = cursor.execute(query)
        columns = [d[0] for d in c.description]
        for row in c:
            yield _Row(values=dict(zip(columns, row)),
                       table_name=self._table_name,
                       con=self._con, header=self._gpheader)
    #----------------------------------------------------------------------
    def _flag_to_bytes(self, code):
        """converts single integer to bytes"""
        return int(code).to_bytes(1,
                                  byteorder='little')
    #----------------------------------------------------------------------
    def _srid_to_bytes(self, srid):
        """converst WKID values to bytes"""
        return int(srid).to_bytes(4,
                                  byteorder='little')
    #----------------------------------------------------------------------
    def _build_gp_header(self, version=0, empty=1):
        """assembles the GP header for WKB geometry"""
        return b'GP' + self._flag_to_bytes(version) + self._flag_to_bytes(empty) + self._srid_to_bytes(self.wkid)
    #----------------------------------------------------------------------
    @property
    def _gpheader(self):
        """internal only, builds the geopackage binary header"""
        if self._gp_header is None:
            self._gp_header = self._build_gp_header()
        return self._gp_header
    #----------------------------------------------------------------------
    def insert(self, row, geom_format='EsriJSON'):
        """
        Inserts a new row via dictionary

        ===============     ===============================================
        **Arguements**      **Description**
        ---------------     -----------------------------------------------
        row                 Required Dictionary.  Insert a new row via a
                            dictionary. The key/value pair must match up to
                            the field names in the table.
        ---------------     -----------------------------------------------
        geom_format         Optional String. When providing geometries
                            during insertion of new records, the method
                            needs to know the format of the geometry. The
                            formats supported values are: EsriJSON,
                            GeoJSON, WKT, and WKB.

                            The default geometry format is`EsriJSON`.

                            **Note**

                            GeoJSON and WKT require the package `geomet` to
                            be installed.

        ==============      ===============================================

        :returns: Boolean

        """
        if _HASGEOMET == False and geom_format.lower() in ['wkt', 'geojson']:
            raise ValueError(("The package `geomet` is required to work with "
                              "WKT and GeoJSON. Run `pip install geomet` to install."))
        if isinstance(row, _Row):
            row = row._values
        values = None
        flds = {fld.lower(): fld for fld in row.keys()}
        if 'shape' in flds:
            if isinstance(row[flds['shape']], dict) and geom_format.lower() == "esrijson":
                row[flds['shape']] = self._gpheader + dumps(row[flds['shape']], False)
            elif isinstance(row[flds['shape']], dict) and geom_format.lower() == "geojson":
                row[flds['shape']] = self._gpheader + geometwkb.dumps(obj=row[flds['shape']])
            elif isinstance(row[flds['shape']], str) and geom_format.lower() == "wkt":
                gj = geometwkt.loads(row[flds['shape']])
                row[flds['shape']] = self._gpheader + geometwkb.dumps(obj=gj)
            elif isinstance(row[flds['shape']], (bytes, bytearray)):
                if isinstance(row[flds['shape']], (bytearray)):
                    row[flds['shape']] = bytes(row[flds['shape']])
                if len(row[flds['shape']]) > 2 and \
                   row[flds['shape']][:2] != b'GB':
                    row[flds['shape']] = self._gpheader + row[flds['shape']]
            elif row[flds['shape']] is None:
                row[flds['shape']] = self._gpheader + b'0x000000000000f87f'
            else:
                raise ValueError(("Shape column must be Esri JSON dictionary, "
                                  "WKT, GeoJSON dictionary, or WKB (bytes)"))
        if isinstance(row, dict):
            keys = row.keys()
            values = [list(row.values())]
        elif isinstance(row, (list, tuple)):
            keys = row[0].keys()
            values = [list(r.values()) for r in row]
        q = ["?"] * len(keys)
        q = ",".join(q)
        sql = '''INSERT INTO {table} ({fields})
                     VALUES({q})'''.format(table=self._table_name,
                                           fields=",".join(keys),
                                           q=q)
        inserts = []
        cur = self._con.cursor()
        cur.execute(sql, list(values[0]))
        self._con.commit()
        return True




"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup
from setuptools import find_packages
from setuptools.command.develop import develop as _develop
from setuptools.command.install import install as _install
from setuptools.command.egg_info import egg_info as _egg_info

# To use a consistent encoding
import os
import sys
from glob import glob
import logging

#TODO
from distutils.core import setup
from codecs import open
from os import path


here = path.abspath(path.dirname(__file__))

packages = ['geopackage']

# Get the long description from the README file
try:
    with open(path.join(here, 'README.md'), encoding='utf-8') as f:
        long_decription = f.read()
except:
    long_decription = "Python GeoPackage Package"

from setuptools import setup


def read_file(file):
    with open(file, 'rb') as fh:
        data = fh.read()
    return data.decode('utf-8')

setup(name='geopackage',
      version='1.0.0',
      description='Pure Python reader/writer of geopackages',
      long_description=read_file('README.md'),
      long_description_content_type='text/markdown',
      author='Andrew Chapkowski',
      author_email='andrewonboe@gmail.com',
      url='https://github.com/achapkowski/pygeopackage',
      py_modules=['geopackage'],
      packages=packages,
      license='Apache License 2.0',
      keywords='gis, geospatial, geographic, geopackage, ogc, wkb, wkt, geojson, spatial, Esri, ArcGIS, Python, ArcPy, qgis',
      python_requires='>= 2.7',
      classifiers=[
                   'Topic :: Scientific/Engineering :: GIS',
                   'Topic :: Software Development :: Libraries',
                   'Topic :: Software Development :: Libraries :: Python Modules',
                   'Topic :: Software Development :: Esri REST API',
                   'Intended Audience :: Developers/GIS Users',
                   'License :: Apache License',
                   'Programming Language :: Python :: 3',
                   'Programming Language :: Python :: 3.2',
                   'Programming Language :: Python :: 3.3',
                   'Programming Language :: Python :: 3.4',
                   'Programming Language :: Python :: 3.5',
                   'Programming Language :: Python :: 3.6',
                   'Development Status :: 5 - Production/Stable'
      ],
      include_package_data=True,
      zip_safe=False,
      install_requires=['pandas'],
      extras_require={},
      package_data={'geopackage' : ['prj.json']}
      )

